from unittest import TestCase

import homotopy.syntax_tree as st


class TestSyntaxTree(TestCase):
    def test_repr(self):
        self.snippet = st.CompositeSnippet(st.SimpleSnippet('if'), '$', st.SimpleSnippet('i==1'))
        self.assertEqual(str(self.snippet), 'if$i==1')

    def test_eq(self):
        # should not be equal to any instance of any other type
        self.assertTrue(st.SimpleSnippet('if') != 123)
        self.assertTrue(st.CompositeSnippet(None, None, None) != 123)

        self.assertEqual(st.SimpleSnippet('if'),
                         st.SimpleSnippet('if'))

        self.assertEqual(st.CompositeSnippet(st.SimpleSnippet('if'), '$', st.SimpleSnippet('i==1')),
                         st.CompositeSnippet(st.SimpleSnippet('if'), '$', st.SimpleSnippet('i==1')))

        self.assertFalse(st.SimpleSnippet('if') != st.SimpleSnippet('if'))

    def test_visitor(self):
        self.assertIsNone(
            st.SnippetVisitor().visit(st.CompositeSnippet(st.SimpleSnippet('if'), '$', st.SimpleSnippet('i==1'))))



import os

from homotopy import preprocessor, parser, compiler, snippet_provider, util


class Homotopy:
    """
    Facade class for accessing homotopy tool.
    """
    stdlib_path = os.path.join(os.path.split(__file__)[0], 'stdlib')

    def __init__(self, language):
        """
        Initialize instance.

        :param language: Language
        """
        self.user_path = []
        self.indent = '\t'
        self.put_cursor_marker = False
        self.language = language

    def add_lib_folder(self, path):
        """
        Add path to a snippet library

        :param path: Path
        """
        self.user_path.append(path)

    def clear_user_lib(self):
        """
        Clear user lib paths.
        """
        self.user_path.clear()

    def set_indent(self, indent):
        """
        Set indent sequence used for indentation instead of tab.

        :param indent: Indent sequence
        """
        self.indent = indent

    def enable_cursor_marker(self):
        """
        Enable cursor marker.
        """
        self.put_cursor_marker = True

    def disable_cursor_marker(self):
        """
        Disable cursor marker.
        """
        self.put_cursor_marker = False

    def set_language(self, language):
        """
        Set language.

        :param language: Language
        """
        self.language = language

    def compile(self, snippet_text):
        """
        Compile a snippet.

        :param snippet_text: Snippet text
        :return: Compiled snippet text
        """
        snippet_provider_instance = snippet_provider.SnippetProvider(
            self.language, self.user_path + [Homotopy.stdlib_path]
        )
        indent_manager_instance = util.IndentManager()

        snippet_text = indent_manager_instance.take_base_indent(snippet_text)

        preprocessor_instance = preprocessor.Preprocessor(snippet_provider_instance)
        compiler_instance = compiler.Compiler(snippet_provider_instance, indent_manager_instance)
        parser_instance = parser.Parser()

        preprocessed_text = preprocessor_instance.expand_decorators(snippet_text)
        if self.put_cursor_marker:
            preprocessed_text = preprocessor_instance.put_cursor_marker(preprocessed_text)

        syntax_tree = parser_instance.parse(preprocessed_text)
        compiled_snippet = compiler_instance.compile(syntax_tree)

        return indent_manager_instance.indent_base(compiled_snippet)



from unittest import TestCase
from unittest.mock import patch, MagicMock
from unittest.mock import mock_open

import os

import homotopy.snippet_provider as sp


class TestSnippetProvider(TestCase):
    def test_getitem(self):
        provider = sp.SnippetProvider("", [])

        provider.data = {
            "for": "for # in !:\n\tpass"
        }

        self.assertEqual(provider['for'], "for # in !:\n\tpass")
        self.assertEqual(provider['i==5'], "i==5")

    @patch('os.listdir')
    def test_initialization(self, listdir):
        listdir.return_value = ['test.json', 'test.txt']

        self.assertDictEqual(sp.SnippetProvider("", []).data, {}, "Snippet provider should be empty.")

        with patch('builtins.open',
                   mock_open(read_data='[{"name": "for","language": "C++","snippet": "if(#){$}"}]')) as m:
            provider = sp.SnippetProvider("c++", ["test"])

            self.assertEqual("if(#){$}", provider["for"])

            m.assert_called_once_with(os.path.join("test", "test.json"))

            listdir.assert_called_once_with("test")

        with patch('builtins.open',
                   mock_open(read_data='[{"name": "for","language": "C++","snippet": "if(#){$}"}]')) as m:
            provider = sp.SnippetProvider("java", ["test"])

            self.assertEqual("for", provider["for"])

        with patch('logging.warning', MagicMock()) as m:
            with patch('builtins.open', mock_open()) as m_open:
                m_open.side_effect = IOError()

                provider = sp.SnippetProvider("c++", ["test"])

                m.assert_called_once_with("Could not get data from file test.json", exc_info=True)

        with patch('logging.warning', MagicMock()) as m:
            with patch('builtins.open', mock_open(read_data='invalidJSON')) as m_open:
                provider = sp.SnippetProvider("c++", ["test"])

                m.assert_called_once_with("Could not get data from file test.json", exc_info=True)

        with patch('logging.warning', MagicMock()) as m:
            with patch('builtins.open', mock_open(
                read_data='[{"name": "for","language": "C++","snippet": "if(#){$}"}' 
                          ',{"name": "for","language": "C++","snippet": "if(#){$}"}]')) as m_open:
                provider = sp.SnippetProvider("c++", ["test"])

                m.assert_called_once_with("Multiple definition for for")

        with patch('builtins.open',
                   mock_open(read_data='[{"name": "for","language": ["C++", "java"],"snippet": "if(#){$}"}]')) as m:
            provider = sp.SnippetProvider("c++", ["test"])

            self.assertEqual("if(#){$}", provider["for"])

        with patch('builtins.open',
                   mock_open(read_data='[{"name": "for","language": ["all"],"snippet": "if(#){$}"}]')) as m:
            provider = sp.SnippetProvider("c++", ["test"])

            self.assertEqual("if(#){$}", provider["for"])

        with patch('builtins.open',
                   mock_open(read_data='[{"name": "for","language": ["~C++"],"snippet": "if(#){$}"}]')) as m:
            provider = sp.SnippetProvider("c++", ["test"])

            self.assertEqual("for", provider["for"])

        with patch('builtins.open',
                   mock_open(read_data='[{"name": "for","language": ["C++", "~C++"],"snippet": "if(#){$}"}]')) as m:
            provider = sp.SnippetProvider("c++", ["test"])

            self.assertEqual("for", provider["for"])

        with patch('builtins.open',
                   mock_open(read_data='[{"name": "for","language": ["all", "~C++"],"snippet": "if(#){$}"}]')) as m:
            provider = sp.SnippetProvider("c++", ["test"])

            self.assertEqual("for", provider["for"])



from test.stdlib_test import TestStdlib


class TestCpp(TestStdlib):
    def setUp(self):
        self.language = "c++"

    def testFlow(self):
        self.verifySingleSnippet("for#int$i%0%n", """
for(int i=0; i<n; i++){

}
[{cursor_marker}]
""")

        self.verifySingleSnippet("for#int$i%0%n>", """
for(int i=0; i<n; i++){
\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet('for#int$i%0%n>printf("hello");', """
for(int i=0; i<n; i++){
\tprintf("hello");
\t[{cursor_marker}]
}
                """)

        self.verifySingleSnippet('if$i==3>return 5;', """
if(i==3){
\treturn 5;
\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet('if$i==2>return 4;<else>return 3;', """
if(i==2){
\treturn 4;
}
else {
\treturn 3;
\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet('while$i==3>return 5;', """
while(i==3){
\treturn 5;
\t[{cursor_marker}]
}

""")

        self.verifySingleSnippet('forr#int$i%n%0>printf("hello");', """
for(int i=n; i>=0; i--){
\tprintf("hello");
\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet('forr#int$i%n%0>printf("hello");<', """
for(int i=n; i>=0; i--){
\tprintf("hello");
}
[{cursor_marker}]
""")


        self.verifySingleSnippet('forin#auto\&\&$item%collection>printf("hello");', """
for(auto&& item: collection){
\tprintf("hello");
\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet('switch$i>case$1>asd<case$2>dsa', """
switch(i){
\tcase 1:
\t\tasd
\t\tbreak;
\tcase 2:
\t\tdsa
\t\t[{cursor_marker}]
\t\tbreak;
}
""")

        self.verifySingleSnippet('switch$i>case$1$2>printf("one or two");', """
switch(i){
\tcase 1:
\tcase 2:
\t\tprintf("one or two");
\t\t[{cursor_marker}]
\t\tbreak;
}
""")

        self.verifySingleSnippet('switch$i>case$1>asd<case$2>dsa<<', """
switch(i){
\tcase 1:
\t\tasd
\t\tbreak;
\tcase 2:
\t\tdsa
\t\tbreak;
}
[{cursor_marker}]
""")

        self.verifySingleSnippet('switch$i>case$1>asd<case$2>dsa<', """
switch(i){
\tcase 1:
\t\tasd
\t\tbreak;
\tcase 2:
\t\tdsa
\t\tbreak;
\t[{cursor_marker}]
}
""")

    def testObjects(self):
        self.verifySingleSnippet('struct!A%public:B>', """
struct A: public B {
\t[{cursor_marker}]
};
""")

        self.verifySingleSnippet('struct!A:B:C%public>', """
struct A: public B, public C {
\t[{cursor_marker}]
};
""")

        self.verifySingleSnippet('class!A%protected:B%public:C>blah', """
class A: protected B, public C {
blah
[{cursor_marker}]
};
""")

        self.verifySingleSnippet('class!A>public>constr#int$i&constr#int$i#int$j', """
class A {
public:
\tA(int i){

\t}

\tA(int i, int j){

\t}

\t[{cursor_marker}]
};
""")

        self.verifySingleSnippet('class!A%public:B>private>constr#int$i', """
class A: public B {
private:
\tA(int i){

\t}

\t[{cursor_marker}]
};
""")

        self.verifySingleSnippet('class!A%public:B>public>econstr#int$i', """
class A: public B {
public:
\tA(int i){}

\t[{cursor_marker}]
};
""")

        self.verifySingleSnippet('class!A%public:B>public>niconstr#int$i', """
class A: public B {
public:
\tA(int i);

\t[{cursor_marker}]
};
""")

        self.verifySingleSnippet('class!A%public:B>private>constr#int$i<public>method#void@test#int$value', """
class A: public B {
private:
\tA(int i){

\t}
public:
\tvoid test(int value){

\t}

\t[{cursor_marker}]
};
""")

        self.verifySingleSnippet('class!A%public:B>private>constr#int$i<public>cmethod#void@test#int$value', """
class A: public B {
private:
\tA(int i){

\t}
public:
\tvoid test(int value) const{

\t}

\t[{cursor_marker}]
};
""")

        self.verifySingleSnippet('class!A%public:B>public>amethod#void@name', """
class A: public B {
public:
\tvirtual void name() = 0;

\t[{cursor_marker}]
};
""")

        self.verifySingleSnippet('class!A%public:B>public>acmethod#void@name', """
class A: public B {
public:
\tvirtual void name() const = 0;

\t[{cursor_marker}]
};
""")

        self.verifySingleSnippet('  class!A%public:B>public>amethod#void@name', """
  class A: public B {
  public:
  \tvirtual void name() = 0;

  \t[{cursor_marker}]
  };
""")

        self.verifySingleSnippet('  class!A%public:B>public>nimethod#void@name', """
  class A: public B {
  public:
  \tvoid name();

  \t[{cursor_marker}]
  };
""")

        self.verifySingleSnippet('  class!A%public:B>public>nicmethod#void@name', """
  class A: public B {
  public:
  \tvoid name() const;

  \t[{cursor_marker}]
  };
""")

        self.verifySingleSnippet('  class!A%public:B>private>nimethod#void@name', """
  class A: public B {
  private:
  \tvoid name();

  \t[{cursor_marker}]
  };
""")

        self.verifySingleSnippet('class!A%public:B>[[singleton]]', """
class A: public B {
public:
\tA& getInstance(){
\t\tstatic A instance;

\t\treturn instance;
\t}
private:
\tA(){}
\tA(A const& origin);
\tvoid operator=(A const& origin);

\t[{cursor_marker}]
};
""")

        self.verifySingleSnippet('class!Composite:Component%public>[[compositeclass]]public>method#void@traverse>[[compositemethod]]<', """
class Composite: public Component {
public:
\tvoid add(Component *item){
\t\tchildren.puch_back(item);
\t}
private:
\tstd::vector<Component*> children;
public:
\tvoid traverse(){
\t\tfor(int i=0; i<children.size(); i++){
\t\t\tchildren[i]->traverse();
\t\t}
\t}

\t[{cursor_marker}]
};
""")

        self.verifySingleSnippet('class!A>public>dconstr#int$i', """
class A {
public:
\tA(int i) = delete;

\t[{cursor_marker}]
};
""")

        self.verifySingleSnippet('class!A>public>dmethod@nothing#void#int$i', """
class A {
public:
\tvoid nothing(int i) = delete;

\t[{cursor_marker}]
};
""")

        self.verifySingleSnippet('class!A>public>dcmethod@nothing#void#int$i', """
class A {
public:
\tvoid nothing(int i) const = delete;

\t[{cursor_marker}]
};
""")

        self.verifySingleSnippet('enum1!deca>pera&mika = 2&zika = 3', """
enum deca { pera, mika = 2, zika = 3, [{cursor_marker}] };
""")

        self.verifySingleSnippet('enum1!deca>pera&mika = 2&zika = 3<', """
enum deca { pera, mika = 2, zika = 3 };
[{cursor_marker}]
""")

        self.verifySingleSnippet('enum!deca>pera&mika = 2&zika = 3<', """
enum deca {
\tpera,
\tmika = 2,
\tzika = 3
};
[{cursor_marker}]
""")

    def testTemplate(self):
        self.verifySingleSnippet('class!A^T', """
template <class T>
class A {

};
[{cursor_marker}]
""")

        self.verifySingleSnippet('func#void@nothing^T', """
template <class T>
void nothing(){

}
[{cursor_marker}]
""")

        self.verifySingleSnippet('class!A>public>method#void@nothing^T', """
class A {
public:
\ttemplate <class T>
\tvoid nothing(){

\t}

\t[{cursor_marker}]
};
""")

    def testMethodImplementation(self):
        self.verifySingleSnippet('constri1!A#int$i', """
A::A(int i){

}
[{cursor_marker}]
""")

        self.verifySingleSnippet('methodi1!A#void@pera#int$i', """
void A::pera(int i){

}
[{cursor_marker}]
""")

        self.verifySingleSnippet('wblock!A>constri&constri$i$j#int&methodi#void@nothing', """
A::A(){

}

A::A(int i, int j){

}

void A::nothing(){

}

[{cursor_marker}]
""")

    def testNesting(self):
        self.verifySingleSnippet('if$one>if$two>if$3', """
if(one){
\tif(two){
\t\tif(3){

\t\t}
\t\t[{cursor_marker}]
\t}
}
""")

    def testMain(self):
        self.verifySingleSnippet('stdinc$stdio.h&[[main]]>printf("Hello, world\!");& &return 0;', """
#include <stdio.h>
int main(int argc, char* argv[]){
\tprintf("Hello, world!");

\treturn 0;
\t[{cursor_marker}]
}
""")


from unittest import TestCase
from unittest.mock import patch, call

import homotopy.syntax_tree as st
from homotopy.compiler import Compiler, ContextManager
from homotopy.snippet_provider import SnippetProvider
from homotopy.util import IndentManager


class TestCompiler(TestCase):
    def setUp(self):
        self.compiler_instance = Compiler(SnippetProvider("", []), IndentManager())

    @patch('homotopy.util.IndentManager.indent_new_lines')
    @patch('homotopy.snippet_provider.SnippetProvider.__getitem__')
    def test_compile(self, mock_provider, mock_indent_new_lines):
        with self.assertRaises(NotImplementedError):
            self.compiler_instance.compile(st.Snippet())

        data = {
            "for": "for ### in !!!:\n\tpass",
            "def": "def !!!({{params}}):\n\tpass",
            "params": "###{{opt_params}}",
            "opt_params": ", ###{{opt_params}}",
            "multiple": "{{goo}}{{doo}}",
            "regextest": "{{{goo}}{{{doo}}}",
            "goo": "goo ###",
            "doo": "doo $$$",
            "outer": "param: ###, inside: >>>",
            "inner": "{{?###}} in: >>>"
        }

        mock_provider.side_effect = lambda x: x if x not in data else data[x]
        mock_indent_new_lines.side_effect = lambda x, _: x

        self.assertEqual(
            self.compiler_instance.compile(st.CompositeSnippet(
                st.CompositeSnippet(st.SimpleSnippet('for'), '#', st.SimpleSnippet('i')),
                '!',
                st.SimpleSnippet('data')
            )),
            'for i in data:\n\tpass'
        )

        self.assertEqual(
            self.compiler_instance.compile(st.CompositeSnippet(
                st.SimpleSnippet('multiple'),
                '$',
                st.SimpleSnippet('asd')
            )),
            'doo asd'
        )

        self.assertEqual(
            self.compiler_instance.compile(st.CompositeSnippet(
                st.SimpleSnippet('regextest'),
                '#',
                st.SimpleSnippet('asd')
            )),
            '{goo asd{}'
        )

        self.assertEqual(
            self.compiler_instance.compile(st.CompositeSnippet(
                st.SimpleSnippet('multiple'),
                '#',
                st.SimpleSnippet('asd')
            )),
            'goo asd'
        )

        with patch('homotopy.compiler.ContextManager.add_variable') as mock_add_variable:
            self.assertEqual(
                self.compiler_instance.compile(st.CompositeSnippet(
                    st.CompositeSnippet(st.SimpleSnippet('for'), '#', st.SimpleSnippet('i')),
                    '%',
                    st.SimpleSnippet('data')
                )),
                'for i in !!!:\n\tpass'
            )

            mock_add_variable.assert_has_calls([
                call('###', 'i'),
                call('%%%', 'data')
            ])

        with patch('homotopy.compiler.ContextManager.add_variable') as mock_add_variable:
            self.assertEqual(
                self.compiler_instance.compile(st.CompositeSnippet(
                    st.CompositeSnippet(
                        st.SimpleSnippet('doo'),
                        '$',
                        st.CompositeSnippet(
                            st.CompositeSnippet(
                                st.SimpleSnippet('def'),
                                '!',
                                st.SimpleSnippet('foo')
                            ),
                            '#',
                            st.SimpleSnippet('a')
                        )),
                    '#',
                    st.SimpleSnippet('b')
                )),
                'doo def foo(a):\n\tpass'
            )

            mock_add_variable.assert_has_calls([
                call('!!!', 'foo'),
                call('###', 'a'),
                call('$$$', 'def foo(a):\n\tpass'),
                call('###', 'b'),
            ])

        self.assertEqual(
            self.compiler_instance.compile(st.CompositeSnippet(
                st.CompositeSnippet(st.SimpleSnippet('def'), '!', st.SimpleSnippet('foo')),
                '#',
                st.SimpleSnippet('a')
            )),
            'def foo(a):\n\tpass'
        )

        self.assertEqual(
            self.compiler_instance.compile(st.CompositeSnippet(
                st.CompositeSnippet(
                    st.CompositeSnippet(st.SimpleSnippet('def'), '!', st.SimpleSnippet('foo')),
                    '#',
                    st.SimpleSnippet('a')
                ),
                '#',
                st.SimpleSnippet('b'))),
            'def foo(a, b):\n\tpass'
        )

        self.assertEqual(
            self.compiler_instance.compile(st.CompositeSnippet(
                st.CompositeSnippet(
                    st.SimpleSnippet('outer'),
                    '#',
                    st.SimpleSnippet('test')
                ),
                '>',
                st.SimpleSnippet('inner'))),
            'param: test, inside: test in: >>>'
        )

        self.assertEqual(
            self.compiler_instance.compile(st.CompositeSnippet(
                st.SimpleSnippet("goo"),
                '#',
                st.SimpleSnippet('goo'))),
            'goo goo'
        )

        self.assertEqual(
            self.compiler_instance.compile(st.CompositeSnippet(
                st.CompositeSnippet(
                    st.SimpleSnippet('outer'),
                    '#',
                    st.SimpleSnippet('test')
                ),
                '>',
                st.CompositeSnippet(
                    st.SimpleSnippet('inner'),
                    '>',
                    st.SimpleSnippet('inner')
                ))),
            'param: test, inside: test in: test in: >>>'
        )

    @patch('homotopy.util.IndentManager.indent_new_lines')
    @patch('homotopy.snippet_provider.SnippetProvider.__getitem__')
    def test_indentation(self, mock_provider, mock_indent_new_lines):
        with self.assertRaises(NotImplementedError):
            self.compiler_instance.compile(st.Snippet())

        data = {
            "for": "for item in collection:\n\t  \t {{helper}}",
            "helper": ">>>",
            "if": "if(true){\n{{inside_block}}\n}",
            "inside_block": "\t>>>{{opt_inside_block}}",
            "opt_inside_block": "\n\t>>>{{opt_inside_block}}"
        }

        mock_provider.side_effect = lambda x: x if x not in data else data[x]
        mock_indent_new_lines.side_effect = lambda x, _: x

        self.assertEqual(
            self.compiler_instance.compile(st.CompositeSnippet(
                st.SimpleSnippet('for'),
                '>',
                st.SimpleSnippet('pass')
            )),
            'for item in collection:\n\t  \t pass'
        )

        mock_indent_new_lines.assert_called_once_with('pass', '\t  \t ')


class TestContextManager(TestCase):
    def setUp(self):
        self.cm = ContextManager()

    def test_init(self):
        self.assertEqual([], self.cm.stack)

    def test_new_scope(self):
        self.cm.new_scope()

        self.assertEqual([{}], self.cm.stack)

    def test_remove_scope(self):
        self.cm.new_scope()
        self.cm.remove_scope()

        self.assertEqual([], self.cm.stack)

        self.cm.remove_scope()
        self.assertEqual([], self.cm.stack)

    def test_add_variable(self):
        with self.assertRaises(Exception) as context:
            self.cm.add_variable("x", "3")

        self.assertEqual("No scope to add variable to", str(context.exception))

        self.cm.new_scope()

        self.cm.add_variable("x", "3")
        self.assertEqual([{"x": "3"}], self.cm.stack)

        self.cm.add_variable("x", "4")
        self.assertEqual([{"x": "4"}], self.cm.stack)

        self.cm.add_variable("y", "3")
        self.assertEqual([{"x": "4", "y": "3"}], self.cm.stack)

    def test_get_item(self):
        self.assertEqual("", self.cm["x"])

        self.cm.new_scope()
        self.cm.add_variable("x", "3")
        self.cm.add_variable("y", "3")

        self.assertEqual("", self.cm["x"])

        self.cm.new_scope()
        self.assertEqual("3", self.cm["x"])

        self.cm.add_variable("x", "4")
        self.assertEqual("3", self.cm["x"])

        self.cm.remove_scope()
        self.assertEqual("", self.cm["x"])



from homotopy.syntax_tree import SimpleSnippet, CompositeSnippet


class Parser:
    """
    Class for parsing a string to produce a syntax tree.
    """
    parameter_chars = "!@#$%:~^"
    in_operator = '>'
    out_operator = '<'
    and_operator = '&'
    escape_character = '\\'

    @staticmethod
    def parse(snippet_text):
        """
        Parsing a string to produce syntax tree.

        :param snippet_text: Text to parse
        :return: syntax tree instance corresponding to given text
        """
        stack = []
        current_match = []
        last_operator = Parser.in_operator
        in_escape_sequence = False

        for c in snippet_text + "\0":
            if in_escape_sequence and c != "\0":
                current_match.append(c)
                in_escape_sequence = False
                continue

            if c == Parser.escape_character:
                in_escape_sequence = True
                continue

            if c in Parser.parameter_chars or c in ["\0", Parser.in_operator, Parser.out_operator, Parser.and_operator]:
                if last_operator == Parser.in_operator:
                    stack.append(SimpleSnippet("".join(current_match)))
                else:
                    current_snippet = stack.pop()
                    stack.append(
                        CompositeSnippet(current_snippet, last_operator, SimpleSnippet("".join(current_match))))

                last_operator = c
                current_match.clear()
            else:
                current_match.append(c)

            if c == Parser.and_operator:
                last_operator = Parser.in_operator
                Parser.merge_stack(stack)

            if c == Parser.out_operator:
                last_operator = Parser.in_operator

                for _ in range(2):
                    Parser.merge_stack(stack)

        while len(stack) > 1:
            Parser.merge_stack(stack)

        return stack.pop()

    @staticmethod
    def merge_stack(stack):
        """
        Helper method to combine two snippets on top of the stack.

        :param stack: Stack
        :return: None
        """
        last_tree = stack.pop()
        if last_tree != SimpleSnippet(''):
            next_tree = stack.pop() if stack else SimpleSnippet("block")
            stack.append(CompositeSnippet(next_tree, Parser.in_operator, last_tree))


parser = Parser()



from test.stdlib_test import TestStdlib


class TestC(TestStdlib):
    def setUp(self):
        self.language = "C"

    def testFlow(self):
        self.verifySingleSnippet("for#int$i%0%n", """
for(int i=0; i<n; i++){

}
[{cursor_marker}]
""")

        self.verifySingleSnippet("for#int$i%0%n>", """
for(int i=0; i<n; i++){
\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet('for#int$i%0%n>printf("hello");', """
for(int i=0; i<n; i++){
\tprintf("hello");
\t[{cursor_marker}]
}
                """)

        self.verifySingleSnippet('if$i==3>return 5;', """
if(i==3){
\treturn 5;
\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet('if$i==2>return 4;<else>return 3;', """
if(i==2){
\treturn 4;
}
else {
\treturn 3;
\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet('while$i==3>return 5;', """
while(i==3){
\treturn 5;
\t[{cursor_marker}]
}

""")

        self.verifySingleSnippet('forr#int$i%n%0>printf("hello");', """
for(int i=n; i>=0; i--){
\tprintf("hello");
\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet('forr#int$i%n%0>printf("hello");<', """
for(int i=n; i>=0; i--){
\tprintf("hello");
}
[{cursor_marker}]
""")

        self.verifySingleSnippet('switch$i>case$1>asd<case$2>dsa', """
switch(i){
\tcase 1:
\t\tasd
\t\tbreak;
\tcase 2:
\t\tdsa
\t\t[{cursor_marker}]
\t\tbreak;
}
""")

        self.verifySingleSnippet('switch$i>case$1$2>printf("one or two");', """
switch(i){
\tcase 1:
\tcase 2:
\t\tprintf("one or two");
\t\t[{cursor_marker}]
\t\tbreak;
}
""")

        self.verifySingleSnippet('switch$i>case$1>asd<case$2>dsa<<', """
switch(i){
\tcase 1:
\t\tasd
\t\tbreak;
\tcase 2:
\t\tdsa
\t\tbreak;
}
[{cursor_marker}]
""")

        self.verifySingleSnippet('switch$i>case$1>asd<case$2>dsa<', """
switch(i){
\tcase 1:
\t\tasd
\t\tbreak;
\tcase 2:
\t\tdsa
\t\tbreak;
\t[{cursor_marker}]
}
""")

    def testObjects(self):
        self.verifySingleSnippet('struct!A>int a;', """
struct A {
\tint a;
\t[{cursor_marker}]
};
""")

        self.verifySingleSnippet('tdstruct!A>int a;&int b;', """
typedef struct{
\tint a;
\tint b;
\t[{cursor_marker}]
} A;
""")

    def testMain(self):
        self.verifySingleSnippet('stdinc$stdio.h&[[main]]>printf("Hello, world\!");& &return 0;', """
#include <stdio.h>
int main(int argc, char* argv[]){
\tprintf("Hello, world!");

\treturn 0;
\t[{cursor_marker}]
}
""")



from unittest import TestCase
from unittest.mock import patch, MagicMock

from homotopy.homotopy import Homotopy


class TestHomotopy(TestCase):
    def setUp(self):
        self.homotopy_instance = Homotopy("c++")

    def test_init(self):
        self.assertEqual([], self.homotopy_instance.user_path)
        self.assertEqual('\t', self.homotopy_instance.indent)
        self.assertEqual(False, self.homotopy_instance.put_cursor_marker)
        self.assertEqual("c++", self.homotopy_instance.language)

    def test_add_lib_folder(self):
        self.homotopy_instance.add_lib_folder("test_folder1")
        self.assertEqual(["test_folder1"], self.homotopy_instance.user_path)

        self.homotopy_instance.add_lib_folder("test_folder2")
        self.assertEqual(["test_folder1", "test_folder2"], self.homotopy_instance.user_path)

    def test_clear_user_lib(self):
        self.homotopy_instance.add_lib_folder("test_folder1")
        self.homotopy_instance.clear_user_lib()
        self.assertEqual([], self.homotopy_instance.user_path)

        self.homotopy_instance.add_lib_folder("test_folder1")
        self.homotopy_instance.add_lib_folder("test_folder2")
        self.homotopy_instance.clear_user_lib()
        self.assertEqual([], self.homotopy_instance.user_path)

    def test_set_indent(self):
        self.homotopy_instance.set_indent("test")
        self.assertEqual("test", self.homotopy_instance.indent)

    def test_enable_cursor_marker(self):
        self.homotopy_instance.enable_cursor_marker()
        self.assertEqual(True, self.homotopy_instance.put_cursor_marker)

    def test_disable_cursor_marker(self):
        self.homotopy_instance.disable_cursor_marker()
        self.assertEqual(False, self.homotopy_instance.put_cursor_marker)

    def test_set_language(self):
        self.homotopy_instance.set_language("python")
        self.assertEqual("python", self.homotopy_instance.language)

    @patch('homotopy.preprocessor.Preprocessor.__init__')
    @patch('homotopy.preprocessor.Preprocessor.expand_decorators')
    @patch('homotopy.preprocessor.Preprocessor.put_cursor_marker')
    @patch('homotopy.parser.Parser.__init__')
    @patch('homotopy.parser.Parser.parse')
    @patch('homotopy.compiler.Compiler.__init__')
    @patch('homotopy.compiler.Compiler.compile')
    @patch('homotopy.snippet_provider.SnippetProvider')
    @patch('homotopy.util.IndentManager')
    def test_compile(
            self,
            mock_indent_manager,
            mock_provider,
            compile_method,
            compile_init,
            parser_parse,
            parser_init,
            preprocessor_put_cursor_marker,
            preprocessor_expand_decorators,
            preprocessor_init,
    ):
        mock_indent_manager_instance = MagicMock()
        mock_indent_manager_instance.take_base_indent = MagicMock(side_effect=lambda x: x)
        mock_indent_manager_instance.indent_base = MagicMock(side_effect=lambda x: x)
        mock_indent_manager.return_value = mock_indent_manager_instance
        mock_provider.return_value = "mock_provider"
        compile_method.return_value = "compile_method_output"
        compile_init.return_value = None
        parser_parse.return_value = "parse_method_output"
        parser_init.return_value = None
        preprocessor_put_cursor_marker.return_value = "put_cursor_marker_output"
        preprocessor_expand_decorators.return_value = "expand_decorators_output"
        preprocessor_init.return_value = None

        self.homotopy_instance.enable_cursor_marker()
        self.homotopy_instance.add_lib_folder("test_folder")

        self.assertEqual("compile_method_output", self.homotopy_instance.compile("test_snippet"))

        mock_provider.assert_called_once_with("c++", ["test_folder", Homotopy.stdlib_path])
        compile_method.assert_called_once_with("parse_method_output")
        compile_init.assert_called_once_with("mock_provider", mock_indent_manager_instance)
        parser_parse.assert_called_once_with("put_cursor_marker_output")
        parser_init.assert_called_once_with()
        preprocessor_put_cursor_marker.assert_called_once_with("expand_decorators_output")
        preprocessor_expand_decorators.assert_called_once_with("test_snippet")
        preprocessor_init.assert_called_once_with("mock_provider")
        mock_indent_manager_instance.take_base_indent.assert_called_once_with("test_snippet")
        mock_indent_manager_instance.indent_base.assert_called_once_with("compile_method_output")



import re


class Preprocessor:
    """
    Prepossess snippet text to enable extra feathers.
    """
    cursor_marker = "[{cursor_marker}]"

    def __init__(self, snippet_provider):
        """
        Initialize preprocessor instance.

        :param snippet_provider: Snippet provider
        """
        self.snippet_provider = snippet_provider

    def expand_decorators(self, snippet_text):
        """
        Expand decorators to enable concise writing of common patterns.

        :param snippet_text: Snippet text
        :return: Expanded snippet text
        """
        return re.sub(
                r'\[\[(.*?)\]\]',
                lambda match_group: self.snippet_provider[match_group.group(1)],
                snippet_text)

    @staticmethod
    def put_cursor_marker(snippet_text):
        """
        Put cursor marker witch should be used by editor plugins for better experience.

        :param snippet_text: Snippet text
        :return: Snippet text with marker at the end
        """
        return snippet_text + "&" + Preprocessor.cursor_marker



from unittest import TestCase
from unittest.mock import patch, call

import homotopy.__main__


class TestMain(TestCase):
    @patch("homotopy.__main__.main")
    def test_init(self, mock_main):
        with patch("homotopy.__main__.__name__", "__main__"):
            homotopy.__main__.init()

        mock_main.assert_called_once_with()

    def test_main(self):
        with patch("homotopy.Homotopy.compile", return_value="") as mock_compile:
            with patch("sys.argv", ["testing", "c++", "invalid_snippet"]):
                homotopy.__main__.main()
            mock_compile.assert_called_once_with("invalid_snippet")

        with patch("homotopy.Homotopy.compile", return_value="") as mock_compile:
            with patch("homotopy.Homotopy.set_indent") as mock_set_indent:
                with patch("sys.argv", ["testing", "-t", "3", "c++", "invalid_snippet"]):
                    homotopy.__main__.main()

                mock_set_indent.assert_called_once_with("   ")
            mock_compile.assert_called_once_with("invalid_snippet")

        with patch("homotopy.Homotopy.compile", return_value="") as mock_compile:
            with patch("homotopy.Homotopy.enable_cursor_marker") as mock_enable_cursor_marker:
                with patch("sys.argv", ["testing", "-c", "c++", "invalid_snippet"]):
                    homotopy.__main__.main()

                mock_enable_cursor_marker.assert_called_once_with()
            mock_compile.assert_called_once_with("invalid_snippet")

        with patch("homotopy.Homotopy.compile", return_value="") as mock_compile:
            with patch("homotopy.Homotopy.add_lib_folder") as mock_add_lib_folder:
                with patch("sys.argv", ["testing", "-p", "user_folder", "c++", "invalid_snippet"]):
                    homotopy.__main__.main()

                mock_add_lib_folder.assert_called_once_with("user_folder")
            mock_compile.assert_called_once_with("invalid_snippet")

        with patch("homotopy.Homotopy.compile", return_value="") as mock_compile:
            with patch("homotopy.Homotopy.add_lib_folder") as mock_add_lib_folder:
                with patch("sys.argv", ["testing", "-p", "user_folder1:user_folder2", "c++", "invalid_snippet"]):
                    homotopy.__main__.main()

                mock_add_lib_folder.assert_has_calls([
                    call("user_folder1"),
                    call("user_folder2")
                ])
            mock_compile.assert_called_once_with("invalid_snippet")



from unittest import TestCase
from unittest.mock import patch

from homotopy.preprocessor import Preprocessor
from homotopy.snippet_provider import SnippetProvider


class TestPreprocessor(TestCase):
    def setUp(self):
        self.preprocessor_instance = Preprocessor(SnippetProvider("", []))

    @patch('homotopy.snippet_provider.SnippetProvider.__getitem__')
    def test_expand_decorators(self, mock_provider):
        data = {
            "def": "expansion",
            "def1": "expansion1",
            "def2": "expansion2"
        }

        mock_provider.side_effect = lambda x: x if x not in data else data[x]

        self.assertEqual("noExpansion", self.preprocessor_instance.expand_decorators("noExpansion"))

        self.assertEqual("test_expansion", self.preprocessor_instance.expand_decorators("test_[[def]]"))

        self.assertEqual("multiple_expansion1_expansions_expansion2",
                         self.preprocessor_instance.expand_decorators("multiple_[[def1]]_expansions_[[def2]]"))

    def test_put_cursor_marker(self):
        cursor_marker = "[{cursor_marker}]"

        self.assertEqual("snippet&" + cursor_marker, self.preprocessor_instance.put_cursor_marker("snippet"))



import argparse

from homotopy.homotopy import Homotopy


def main():
    parser = argparse.ArgumentParser(description="Compile a snippet.")
    parser.add_argument('language', type=str,
                        help='Language for the snippet to be compiled to')
    parser.add_argument('snippet', type=str,
                        help='A snippet to be compiled')
    parser.add_argument('-t', '--tabsize', type=int, metavar="N",
                        help='Number of spaces in one tab. Tabs remain tabs if absent')
    parser.add_argument('-c', '--cursor', action='store_true',
                        help='Indicate cursor marker in compiled snippet')
    parser.add_argument('-p', '--path', type=str, metavar="PATH",
                        help='Path to snippet library folders separated by :')

    args = parser.parse_args()

    homotopy = Homotopy(args.language)

    if args.tabsize:
        homotopy.set_indent(" "*args.tabsize)

    if args.cursor:
        homotopy.enable_cursor_marker()

    if args.path:
        for item in args.path.split(':'):
            homotopy.add_lib_folder(item)

    print(homotopy.compile(args.snippet), end='')


def init():
    if __name__ == "__main__":
        main()


init()



from unittest import TestCase

from homotopy.parser import parser
from homotopy.syntax_tree import SimpleSnippet, CompositeSnippet


class TestParser(TestCase):
    def test_basic(self):
        self.assertEqual(parser.parse(''), SimpleSnippet(''))

        self.assertEqual(parser.parse('asd'), SimpleSnippet('asd'))
        self.assertEqual(parser.parse('asd#'), CompositeSnippet(SimpleSnippet('asd'), '#', SimpleSnippet('')))
        self.assertEqual(parser.parse('asd!#1'),
                         CompositeSnippet(
                             CompositeSnippet(SimpleSnippet('asd'), '!',
                                              SimpleSnippet('')), '#',
                             SimpleSnippet('1'))
                         )

    def test_parameters(self):
        operators = '!@#$%:~^'
        for l in operators:
            self.assertEqual(parser.parse('first{0}second{0}third'.format(l)),
                             CompositeSnippet(
                                 CompositeSnippet(SimpleSnippet('first'), l,
                                                  SimpleSnippet('second')), l,
                                 SimpleSnippet('third'))
                             )

    def test_into(self):
        self.assertEqual(parser.parse('asd>dsa'), CompositeSnippet(SimpleSnippet('asd'), '>', SimpleSnippet('dsa')))
        self.assertEqual(parser.parse('asd>dsa#2'),
                         CompositeSnippet(SimpleSnippet('asd'), '>',
                                          CompositeSnippet(SimpleSnippet('dsa'), '#',
                                                           SimpleSnippet('2')))
                         )

        self.assertEqual(parser.parse('for>if&if'),
                         CompositeSnippet(
                             CompositeSnippet(SimpleSnippet('for'), '>',
                                              SimpleSnippet('if')), '>',
                             SimpleSnippet('if'))
                         )

        self.assertEqual(parser.parse('for>if>if<if'),
                         CompositeSnippet(
                             CompositeSnippet(SimpleSnippet('for'), '>',
                                              CompositeSnippet(SimpleSnippet('if'), '>', SimpleSnippet('if'))), '>',
                             SimpleSnippet('if'))
                         )

        self.assertEqual(parser.parse('for>if>if>if<<if'),
                         CompositeSnippet(
                             CompositeSnippet(SimpleSnippet('for'), '>',
                                              CompositeSnippet(SimpleSnippet('if'), '>',
                                                               CompositeSnippet(
                                                                   SimpleSnippet('if'), '>',
                                                               SimpleSnippet('if')))), '>',
                             SimpleSnippet('if'))
                         )

        self.assertEqual(parser.parse('for>if#5>if<if'),
                         CompositeSnippet(
                             CompositeSnippet(SimpleSnippet('for'), '>',
                                              CompositeSnippet(
                                                  CompositeSnippet(SimpleSnippet('if'), '#', SimpleSnippet('5')), '>',
                                                  SimpleSnippet('if'))), '>',
                             SimpleSnippet('if'))
                         )

        self.assertEqual(parser.parse('for&for'),
                         CompositeSnippet(
                             CompositeSnippet(
                                 SimpleSnippet('block'),
                                 '>',
                                 SimpleSnippet('for')),
                             '>',
                             SimpleSnippet('for')
                         ))

        self.assertEqual(parser.parse('for>if<if'),
                         CompositeSnippet(
                             CompositeSnippet(
                                 SimpleSnippet('block'),
                                 '>',
                                 CompositeSnippet(
                                     SimpleSnippet('for'),
                                     '>',
                                     SimpleSnippet('if')
                                 )
                             ),
                             '>',
                             SimpleSnippet('if')
                         ))

    def test_escape(self):
        self.assertEqual(parser.parse(r"i\>4"), SimpleSnippet("i>4"))
        self.assertEqual(parser.parse("ignore\\"), SimpleSnippet("ignore"))

        self.assertEqual(parser.parse(r"i\\>4"),
                         CompositeSnippet(
                             SimpleSnippet("i\\"),
                             '>',
                             SimpleSnippet("4")
                         )
                         )

        self.assertEqual(parser.parse(r"if$i\>4"),
                         CompositeSnippet(
                             SimpleSnippet("if"),
                             '$',
                             SimpleSnippet("i>4")
                         ))





class Snippet:
    """
    Base class for snippet syntax tree.
    """
    def accept(self, visitor):
        """
        Accepts a visitor

        :param visitor: Visitor instance
        :return: Visitor result
        """
        raise NotImplementedError("You should not be here.")


class CompositeSnippet(Snippet):
    """
    CompositeSnippet compose two snippets with the operand
    """
    def __init__(self, left, operation, right):
        """
        Initialize CompositeSnippet instance.

        :param left: Left subtree
        :param operation: Operation
        :param right: Right subtree
        """
        self.left = left
        self.operation = operation
        self.right = right

    def __repr__(self):
        return str(self.left) + self.operation + str(self.right)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.left == other.left and \
                   self.operation == other.operation and\
                   self.right == other.right
        return False

    def accept(self, visitor):
        return visitor.visit_composite_snippet(self)


class SimpleSnippet(Snippet):
    """
    BasicSnippet can be directly compiled.
    """
    def __init__(self, value):
        """
        Initialize SimpleSnippet instance.

        :param value: Value of the snippet
        """
        self.value = value

    def __repr__(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.value == other.value
        return False

    def accept(self, visitor):
        return visitor.visit_simple_snippet(self)


class SnippetVisitor:
    """
    Base class for visitors working on syntax tree.
    """
    def visit(self, snippet):
        return snippet.accept(self)

    def visit_composite_snippet(self, composite_snippet):
        """
        Base visit logic for composite snippets. Process right and left subtrees recursively.

        :param composite_snippet: CompositeSnippet instance
        :return: None
        """
        composite_snippet.left.accept(self)
        composite_snippet.right.accept(self)

    def visit_simple_snippet(self, simple_snippet):
        """
        Base visit logic for simple snippets. Does nothing.

        :param simple_snippet: SimpleSnippet instance
        :return: None
        """
        pass



from distutils.core import setup
import os

setup(
    name='homotopy',
    version='0.1.dev1',
    packages=['homotopy'],
    license='MIT',
    author='Nenad Vasic',
    author_email='ahhhhmed@gmail.com',
    url='https://github.com/Ahhhhmed/homotopy',
    description='Homotopy snippet engine',
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.rst')).read(),
    install_requires=[],
    python_requires='>=3.3',
    include_package_data=True,
    package_data={'homotopy': ['stdlib/*']},
    entry_points={
        'console_scripts': [
            'homotopy = homotopy.__main__:main'
        ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development',
        'Topic :: Software Development :: Compilers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: C',
        'Programming Language :: C++',
        'Programming Language :: Java',
    ],
    test_suite='test',
)



import re


class IndentManager:
    """
    Helper for managing indent.
    """
    def __init__(self, base_indent=""):
        """
        Initializer.

        :param base_indent: Base indent
        """
        self.base_indent = base_indent

    @staticmethod
    def indent_new_lines(text, indent):
        """
        Indent all the lines except first one.

        :param text: Text to indent
        :param indent: Indent text
        :return: Indented text
        """
        lines = text.splitlines()
        lines[1:] = [indent + x if x.strip() != "" else "" for x in lines[1:]]
        return "\n".join(lines)

    def indent_base(self, text):
        """
        Indent all lines with base indent.

        :param text: Text to indent
        :return: Indented text
        """
        lines = text.splitlines()
        lines = [self.base_indent + x if x.strip() != "" else "" for x in lines]
        return "\n".join(lines)

    def take_base_indent(self, line):
        """
        Take base indent.

        :param line: Line text
        :return: Line without the indent
        """
        m = re.match(r"^[\t ]*", line)
        self.base_indent = m.group()

        return line[m.end():]


class ContextManager:
    """
    Class for managing substituted parameters to be used again.
    """
    def __init__(self):
        """
        Initialize stack.
        """
        self.stack = []

    def new_scope(self):
        """
        Add new scope.
        """
        self.stack.append({})

    def remove_scope(self):
        """
        Remove last scope.
        """
        if self.stack:
            self.stack.pop()

    def add_variable(self, name, value):
        """
        Add variable to current scope.

        :param name: Variable name
        :param value: Variable value
        """
        if self.stack:
            self.stack[-1][name] = value
        else:
            raise Exception("No scope to add variable to")

    def __getitem__(self, item):
        """
        Get variable from last scope.

        :param item: Variable name
        :return: Variable value if it exists. Empty string otherwise
        """
        i = 2
        while i <= len(self.stack):
            if item in self.stack[-i]:
                return self.stack[-i][item]

            i += 1

        return ""



import os
import json
import logging


class SnippetProvider:
    """
    Class for translating simple snippets into code.
    Uses a database provided in json files located in the path variable.
    """

    language_key = "language"
    name_key = "name"
    snippet_key = "snippet"

    def __init__(self, language, path):
        """
        Initialize snippet provider instance.

        :param language: language to compile to
        :param path: list of directories to search for json files containing snippets
        """
        self.data = {}
        self.language = language
        self.path = path
        for item in self.path:
            for file_name in filter(lambda x: x.endswith(".json"), os.listdir(item)):
                try:
                    def language_filter(filter_item):
                        all_languages = 'all'

                        languages = [filter_item[SnippetProvider.language_key].lower()] \
                            if type(filter_item[SnippetProvider.language_key]) is str \
                            else [x.lower() for x in filter_item[SnippetProvider.language_key]]

                        if '~' + self.language.lower() in languages:
                            return False

                        return all_languages in languages or self.language.lower() in languages

                    with open(os.path.join(item, file_name)) as opened_file:
                        for snippet in filter(language_filter, json.load(opened_file)):
                            if snippet[SnippetProvider.name_key] in self.data:
                                logging.warning("Multiple definition for %s" % snippet[SnippetProvider.name_key])
                            else:
                                self.data[snippet[SnippetProvider.name_key]] = snippet[SnippetProvider.snippet_key]
                except (ValueError, OSError):
                    logging.warning("Could not get data from file %s" % file_name, exc_info=True)

    def __getitem__(self, item):
        """
        Expand single snippet.

        :param item: snippet
        :return: snippet expansion
        """
        if item in self.data:
            return self.data[item]
        return item



from unittest import TestCase

from homotopy.util import ContextManager, IndentManager


class TestIndentManager(TestCase):
    def setUp(self):
        self.indent_manager_instance = IndentManager()

    def test_init(self):
        self.assertEqual("", self.indent_manager_instance.base_indent)

        self.assertEqual("indent", IndentManager("indent").base_indent)

    def test_indent_new_lines(self):
        self.assertEqual("test", self.indent_manager_instance.indent_new_lines("test", "indent "))

        self.assertEqual("line1\nindent line2",
                         self.indent_manager_instance.indent_new_lines("line1\nline2", "indent "))

        self.assertEqual(
            "line1\nindent line2\nindent line3",
            self.indent_manager_instance.indent_new_lines("line1\nline2\nline3", "indent "))

        self.assertEqual("line1\nindent line2",
                         self.indent_manager_instance.indent_new_lines("line1\nline2\n", "indent "))

        self.assertEqual("line1\nindent line2\n",
                         self.indent_manager_instance.indent_new_lines("line1\nline2\n\n", "indent "))

        self.assertEqual("line1\nindent line2\n",
                         self.indent_manager_instance.indent_new_lines("line1\nline2\n  \n", "indent "))

        self.assertEqual(
            "\nindent line1\nindent line2",
            self.indent_manager_instance.indent_new_lines("\nline1\nline2", "indent "))

    def test_indent_base(self):
        self.indent_manager_instance.base_indent = "indent "

        self.assertEqual("indent test", self.indent_manager_instance.indent_base("test"))
        self.assertEqual("indent line1\nindent line2", self.indent_manager_instance.indent_base("line1\nline2"))

        self.assertEqual(
            "indent line1\nindent line2\nindent line3",
            self.indent_manager_instance.indent_base("line1\nline2\nline3"))

        self.assertEqual("indent line1\nindent line2", self.indent_manager_instance.indent_base("line1\nline2\n"))
        self.assertEqual("indent line1\nindent line2\n", self.indent_manager_instance.indent_base("line1\nline2\n\n"))
        self.assertEqual("indent line1\nindent line2\n", self.indent_manager_instance.indent_base("line1\nline2\n  \n"))

        self.assertEqual(
            "\nindent line1\nindent line2",
            self.indent_manager_instance.indent_base("\nline1\nline2"))

    def test_take_base_indent(self):
        self.assertEqual("", self.indent_manager_instance.take_base_indent("   "))
        self.assertEqual("   ", self.indent_manager_instance.base_indent)

        self.assertEqual("for", self.indent_manager_instance.take_base_indent("   for"))
        self.assertEqual("   ", self.indent_manager_instance.base_indent)

        self.assertEqual("for", self.indent_manager_instance.take_base_indent("\tfor"))
        self.assertEqual("\t", self.indent_manager_instance.base_indent)

        self.assertEqual("for", self.indent_manager_instance.take_base_indent("\t  \t for"))
        self.assertEqual("\t  \t ", self.indent_manager_instance.base_indent)


class TestContextManager(TestCase):
    def setUp(self):
        self.cm = ContextManager()

    def test_init(self):
        self.assertEqual([], self.cm.stack)

    def test_new_scope(self):
        self.cm.new_scope()

        self.assertEqual([{}], self.cm.stack)

    def test_remove_scope(self):
        self.cm.new_scope()
        self.cm.remove_scope()

        self.assertEqual([], self.cm.stack)

        self.cm.remove_scope()
        self.assertEqual([], self.cm.stack)

    def test_add_variable(self):
        with self.assertRaises(Exception) as context:
            self.cm.add_variable("x", "3")

        self.assertEqual("No scope to add variable to", str(context.exception))

        self.cm.new_scope()

        self.cm.add_variable("x", "3")
        self.assertEqual([{"x": "3"}], self.cm.stack)

        self.cm.add_variable("x", "4")
        self.assertEqual([{"x": "4"}], self.cm.stack)

        self.cm.add_variable("y", "3")
        self.assertEqual([{"x": "4", "y": "3"}], self.cm.stack)

    def test_get_item(self):
        self.assertEqual("", self.cm["x"])

        self.cm.new_scope()
        self.cm.add_variable("x", "3")
        self.cm.add_variable("y", "3")

        self.assertEqual("", self.cm["x"])

        self.cm.new_scope()
        self.assertEqual("3", self.cm["x"])

        self.cm.add_variable("x", "4")
        self.assertEqual("3", self.cm["x"])

        self.cm.remove_scope()
        self.assertEqual("", self.cm["x"])



import re

from homotopy.syntax_tree import SnippetVisitor
from homotopy.parser import Parser
from homotopy.util import ContextManager


class Compiler(SnippetVisitor):
    """
    Compiler for snippets. Turns syntax tree into text.
    """
    def __init__(self, snippet_provider, indent_manager):
        self.context_manager = ContextManager()
        self.context_manager.new_scope()
        self.snippet_provider = snippet_provider
        self.indent_manager = indent_manager
        self.inside_parameter = False

    def visit_composite_snippet(self, composite_snippet):
        """
        Replace right site in the appropriate place in the left side.
        Expand left side if needed. Manage scopes.

        :param composite_snippet: Composite snippet
        :return: Text of left side replaced with right side
        """
        left_side = self.expand_variable_operators(self.snippet_provider[self.visit(composite_snippet.left)])
        operation_text = composite_snippet.operation * 3

        if operation_text in left_side:
            return self.substitute(left_side, composite_snippet.operation, composite_snippet.right, operation_text)

        expanded_left_side = self.expand_snippet(left_side, operation_text)
        return self.substitute(expanded_left_side, composite_snippet.operation, composite_snippet.right, operation_text)

    def visit_simple_snippet(self, simple_snippet):
        """
        Compile simple snippet.

        :param simple_snippet: Simple snippet
        :return: Text of compile snippet
        """
        snippet_text = simple_snippet.value if self.inside_parameter else self.snippet_provider[simple_snippet.value]

        return self.expand_variable_operators(snippet_text)

    def expand_snippet(self, snippet_text, operation_text):
        """
        Expend snippet to uncover possible operator definition.

        :param snippet_text: Snippet text
        :param operation_text: Operation text
        :return: Expanded snippet
        """
        match_found = False

        def expansion_function(match_object):
            nonlocal match_found

            if not match_found and operation_text in self.snippet_provider[match_object.group(1)]:
                match_found = True
                return self.snippet_provider[match_object.group(1)]

            return match_object.group(0)

        return re.sub(
            r'{{([^{]*?)}}',
            expansion_function,
            snippet_text)

    def expand_variable_operators(self, text):
        """
        Expend parameter variables.

        :param text: Text
        :return: Expended text
        """
        return re.sub(
            r'{{[?](.*?)}}',
            lambda match_group: self.context_manager[match_group.group(1)],
            text)

    def substitute(self, left, operation, right_tree, operation_text):
        if operation == Parser.in_operator:
            self.context_manager.new_scope()

            before_operation_text = left[0:left.find(operation_text)]
            m = re.search(r"\n([\t ]*)$", before_operation_text)
            indent = m.group(1) if m else ""
        else:
            old_inside_parameter = self.inside_parameter
            self.inside_parameter = True

        right = self.compile(right_tree)

        if operation == Parser.in_operator:
            self.context_manager.remove_scope()

            right = self.indent_manager.indent_new_lines(right, indent)
        else:
            self.context_manager.add_variable(operation_text, right)

            self.inside_parameter = old_inside_parameter

        return left.replace(operation_text, right)

    def compile(self, snippet):
        """
        Compile a snippet. Visit and then perform a clean.

        :param snippet: Snippet
        :return: Text of compiled snippet
        """
        compiled_snippet = self.visit(snippet)

        return re.sub(r'({{[^{]*?}})', "", compiled_snippet)



from test.stdlib_test import TestStdlib


class TestC(TestStdlib):
    def setUp(self):
        self.language = "Java"

    def testFlow(self):
        self.verifySingleSnippet("forin#int$i%collection", """
for(int i: collection){

}
[{cursor_marker}]
""")

    def testObject(self):
        self.verifySingleSnippet("class!A:B~C~D", """
class A extends B implements C, D {

}
[{cursor_marker}]
""")

        self.verifySingleSnippet("class!A>method#void@nothing", """
class A {
\tpublic void nothing(){

\t}

\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet("class!A>method1#void@nothing", """
class A {
\tpublic void nothing(){ }

\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet("class!A>method1#int@five>return 5;", """
class A {
\tpublic int five(){ return 5; [{cursor_marker}] }
}
""")

        self.verifySingleSnippet("class!A>amethod#int@five", """
class A {
\tpublic abstract int five();

\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet("class!A>pmethod#void@nothing", """
class A {
\tprivate void nothing(){

\t}

\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet("class!A>pmethod1#void@nothing", """
class A {
\tprivate void nothing(){ }

\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet("class!A>pmethod1#int@five>return 5;", """
class A {
\tprivate int five(){ return 5; [{cursor_marker}] }
}
""")

        self.verifySingleSnippet("class!A>pamethod#int@five", """
class A {
\tprivate abstract int five();

\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet("class!A>promethod#void@nothing", """
class A {
\tprotected void nothing(){

\t}

\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet("class!A>promethod1#void@nothing", """
class A {
\tprotected void nothing(){ }

\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet("class!A>promethod1#int@five>return 5;", """
class A {
\tprotected int five(){ return 5; [{cursor_marker}] }
}
""")

        self.verifySingleSnippet("class!A>proamethod#int@five", """
class A {
\tprotected int abstract five();

\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet("class!A>constr#int$i", """
class A {
\tpublic A(int i){

\t}

\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet("class!A>constr1#int$i", """
class A {
\tpublic A(int i){ }

\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet("class!A>pconstr#int$i", """
class A {
\tprivate A(int i){

\t}

\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet("class!A>pconstr1#int$i", """
class A {
\tprivate A(int i){ }

\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet("class!A>proconstr#int$i", """
class A {
\tprotected A(int i){

\t}

\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet("class!A>proconstr1#int$i", """
class A {
\tprotected A(int i){ }

\t[{cursor_marker}]
}
""")

        self.verifySingleSnippet("enum1!Color>red&green&blue<", """
enum Color { red, green, blue; }
[{cursor_marker}]
""")

    def testTemplate(self):
        self.verifySingleSnippet("class!A^T>method#void@nothing^K", """
class A<T> {
\tpublic void nothing<K>(){

\t}

\t[{cursor_marker}]
}
""")

    def testPatterns(self):

        self.verifySingleSnippet("class!A:B>[[compositeclass]]&method#void@nothing>[[compositemethod]]<", """
class A extends B {
\tprivate List<B> children = ArrayList<B>();
\tpublic void add(B item){
\t\tchildren.add(item);
\t}

\tpublic void nothing(){
\t\tfor(int i=0; i<children.size(); i++){
\t\t\tchildren[i].nothing();
\t\t}
\t}

\t[{cursor_marker}]
}
""")

    def testMain(self):
        self.verifySingleSnippet("class!A>[[main]]>return 0;<", """
class A {
\tpublic static void main(String[] args){
\t\treturn 0;
\t}

\t[{cursor_marker}]
}
""")





from test.stdlib_test import TestStdlib


class TestC(TestStdlib):
    def setUp(self):
        self.language = "Python"

    def testFlow(self):
        self.verifySingleSnippet("forin$i%collection>pass", """
for i in collection:
\tpass
\t[{cursor_marker}]
""")

        self.verifySingleSnippet("for$i%collection>pass", """
for i in collection:
\tpass
\t[{cursor_marker}]
""")

        self.verifySingleSnippet("if$i is None>pass<elif$i==0>return 3<else>return 2", """
if i is None:
\tpass
elif i==0:
\treturn 3
else:
\treturn 2
\t[{cursor_marker}]
""")

        self.verifySingleSnippet("while$True>pass", """
while True:
\tpass
\t[{cursor_marker}]
""")

    def testObject(self):
        self.verifySingleSnippet("class!A:B>pass", """
class A(B):
\tpass

\t[{cursor_marker}]
""")

        self.verifySingleSnippet("class!A>method@foo$i>pass", """
class A:
\tdef foo(self, i):
\t\tpass
\t\t[{cursor_marker}]
""")

        self.verifySingleSnippet("class!A>smethod@foo$i>pass", """
class A:
\t@staticmethod
\tdef foo(i):
\t\tpass
\t\t[{cursor_marker}]
""")

        self.verifySingleSnippet("class!A>cmethod@foo$i>pass", """
class A:
\t@classmethod
\tdef foo(cls, i):
\t\tpass
\t\t[{cursor_marker}]
""")

        self.verifySingleSnippet("class!A>constr$i>pass", """
class A:
\tdef __init__(self, i):
\t\tpass
\t\t[{cursor_marker}]
""")

        self.verifySingleSnippet("func@a$i>pass", """
def a(i):
\tpass
\t[{cursor_marker}]
""")

    def testMain(self):
        self.verifySingleSnippet("[[main]]>pass", """
if __name__ == "__main__":
	pass
\t[{cursor_marker}]
""")
